import os
from unittest import mock

import pytest

from cloudinit import gpg, subp
from cloudinit.subp import SubpResult

TEST_KEY_HUMAN = """
/etc/apt/cloud-init.gpg.d/my_key.gpg
--------------------------------------------
pub   rsa4096 2021-10-22 [SC]
      3A3E F34D FDED B3B7 F3FD  F603 F83F 7712 9A5E BD85
uid           [ unknown] Brett Holman <brett.holman@canonical.com>
sub   rsa4096 2021-10-22 [A]
sub   rsa4096 2021-10-22 [E]
"""

TEST_KEY_MACHINE = """
tru::1:1635129362:0:3:1:5
pub:-:4096:1:F83F77129A5EBD85:1634912922:::-:::scESCA::::::23::0:
fpr:::::::::3A3EF34DFDEDB3B7F3FDF603F83F77129A5EBD85:
uid:-::::1634912922::64F1F1D6FA96316752D635D7C6406C52C40713C7::Brett Holman \
<brett.holman@canonical.com>::::::::::0:
sub:-:4096:1:544B39C9A9141F04:1634912922::::::a::::::23:
fpr:::::::::8BD901490D6EC986D03D6F0D544B39C9A9141F04:
sub:-:4096:1:F45D9443F0A87092:1634912922::::::e::::::23:
fpr:::::::::8CCCB332317324F030A45B19F45D9443F0A87092:
"""

TEST_KEY_FINGERPRINT_HUMAN = (
    "3A3E F34D FDED B3B7 F3FD  F603 F83F 7712 9A5E BD85"
)

TEST_KEY_FINGERPRINT_MACHINE = "3A3EF34DFDEDB3B7F3FDF603F83F77129A5EBD85"


@pytest.fixture()
def m_subp():
    with mock.patch.object(
        gpg.subp, "subp", return_value=SubpResult("", "")
    ) as m_subp, mock.patch.object(gpg.time, "sleep"):
        yield m_subp


@pytest.fixture()
def m_which():
    with mock.patch.object(gpg.subp, "which") as m_which:
        yield m_which


@pytest.fixture()
def m_sleep():
    with mock.patch("cloudinit.gpg.time.sleep") as sleep:
        yield sleep


class TestGPGCommands:
    def test_dearmor_bad_value(self):
        """This exception is handled by the callee. Ensure it is not caught
        internally.
        """
        gpg_instance = gpg.GPG()
        with mock.patch.object(
            subp, "subp", side_effect=subp.ProcessExecutionError
        ):
            with pytest.raises(subp.ProcessExecutionError):
                gpg_instance.dearmor("garbage key value")

    def test_gpg_list_args(self, m_subp):
        """Verify correct command gets called to list keys"""
        gpg_instance = gpg.GPG()
        no_colons = [
            "gpg",
            "--no-options",
            "--with-fingerprint",
            "--no-default-keyring",
            "--list-keys",
            "--keyring",
            "key",
        ]
        colons = [
            "gpg",
            "--no-options",
            "--with-fingerprint",
            "--no-default-keyring",
            "--list-keys",
            "--keyring",
            "--with-colons",
            "key",
        ]
        gpg_instance.list_keys("key")
        assert (
            mock.call(colons, capture=True, update_env=gpg_instance.env)
            == m_subp.call_args
        )

        gpg_instance = gpg.GPG()
        gpg_instance.list_keys("key", human_output=True)
        assert m_subp.call_args == mock.call(
            no_colons, capture=True, update_env=gpg_instance.env
        )

    def test_gpg_dearmor_args(self, m_subp):
        """Verify correct command gets called to dearmor keys"""
        gpg_instance = gpg.GPG()
        gpg_instance.dearmor("key")
        test_call = mock.call(
            ["gpg", "--dearmor"],
            data="key",
            decode=False,
            update_env=gpg_instance.env,
        )
        assert test_call == m_subp.call_args


class TestReceiveKeys:
    """Test the recv_key method."""

    def test_retries_on_subp_exc(self, m_subp, m_sleep):
        """retry should be done on gpg receive keys failure."""
        gpg_instance = gpg.GPG()
        retries = (1, 2, 4)
        my_exc = subp.ProcessExecutionError(
            stdout="", stderr="", exit_code=2, cmd=["mycmd"]
        )
        m_subp.side_effect = (my_exc, my_exc, ("", ""))
        gpg_instance.recv_key("ABCD", "keyserver.example.com", retries=retries)
        assert [mock.call(1), mock.call(2)], m_sleep.call_args_list

    def test_raises_error_after_retries(self, m_subp, m_sleep):
        """If the final run fails, error should be raised."""
        gpg_instance = gpg.GPG()
        naplen = 1
        keyid, keyserver = ("ABCD", "keyserver.example.com")
        m_subp.side_effect = subp.ProcessExecutionError(
            stdout="", stderr="", exit_code=2, cmd=["mycmd"]
        )
        with pytest.raises(
            ValueError, match=f"{keyid}.*{keyserver}|{keyserver}.*{keyid}"
        ):
            gpg_instance.recv_key(keyid, keyserver, retries=(naplen,))
        m_sleep.assert_called_once()

    def test_no_retries_on_none(self, m_subp, m_sleep):
        """retry should not be done if retries is None."""
        gpg_instance = gpg.GPG()
        m_subp.side_effect = subp.ProcessExecutionError(
            stdout="", stderr="", exit_code=2, cmd=["mycmd"]
        )
        with pytest.raises(ValueError):
            gpg_instance.recv_key(
                "ABCD", "keyserver.example.com", retries=None
            )
        m_sleep.assert_not_called()

    def test_expected_gpg_command(self, m_subp, m_sleep):
        """Verify gpg is called with expected args."""
        gpg_instance = gpg.GPG()
        key, keyserver = ("DEADBEEF", "keyserver.example.com")
        retries = (1, 2, 4)
        m_subp.return_value = ("", "")
        gpg_instance.recv_key(key, keyserver, retries=retries)
        m_subp.assert_called_once_with(
            [
                "gpg",
                "--no-tty",
                "--keyserver=%s" % keyserver,
                "--recv-keys",
                key,
            ],
            capture=True,
            update_env=gpg_instance.env,
        )
        m_sleep.assert_not_called()

    def test_kill_gpg_succeeds(self, m_subp, m_which):
        """ensure that when gpgconf isn't found, processes are manually
        cleaned up. Also test that the context manager does cleanup

        """
        m_which.return_value = True
        with pytest.raises(ZeroDivisionError):
            with gpg.GPG() as gpg_context:

                # run a gpg command so that we have "started" gpg
                gpg_context.list_keys("")
                1 / 0  # pylint: disable=pointless-statement
        m_subp.assert_has_calls(
            [
                mock.call(
                    ["gpgconf", "--kill", "all"],
                    capture=True,
                    update_env=gpg_context.env,
                )
            ]
        )
        assert not os.path.isdir(str(gpg_context.temp_dir))

    def test_do_not_kill_unstarted_gpg(self, m_subp):
        """ensure that when gpg isn't started, gpg isn't killed, but the
        directory is cleaned up.
        """
        with pytest.raises(ZeroDivisionError):
            with gpg.GPG() as gpg_context:
                1 / 0  # pylint: disable=pointless-statement
        m_subp.assert_not_called()
        assert not os.path.isdir(str(gpg_context.temp_dir))

    def test_kill_gpg_failover_succeeds(self, m_subp, m_which):
        """ensure that when gpgconf isn't found, processes are manually
        cleaned up
        """
        m_which.return_value = None
        gpg_instance = gpg.GPG()

        # "start" gpg (if we don't, we won't kill gpg)
        gpg_instance.recv_key("", "")
        gpg_instance.kill_gpg()
        m_subp.assert_has_calls(
            [
                mock.call(
                    [
                        "ps",
                        "-o",
                        "ppid,pid",
                        "-C",
                        "keyboxd",
                        "-C",
                        "dirmngr",
                        "-C",
                        "gpg-agent",
                    ],
                    capture=True,
                    rcs=[0, 1],
                )
            ]
        )



# This file is part of cloud-init. See LICENSE file for license information.

import pytest

from tests.integration_tests import releases
from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.releases import IS_UBUNTU
from tests.integration_tests.util import ASSETS_DIR


@pytest.mark.skipif(
    not IS_UBUNTU, reason="module dir tested is ubuntu-specific"
)
def test_custom_module_24_1(client: IntegrationInstance):
    """Ensure that modifications to cloud-init don't break old custom modules.

    24.1 had documentation that differs from current best practices. We want
    to ensure modules created from this documentation still work:
    https://docs.cloud-init.io/en/24.1/development/module_creation.html
    """
    client.push_file(
        ASSETS_DIR / "dropins/cc_custom_module_24_1.py",
        "/usr/lib/python3/dist-packages/cloudinit/config/cc_custom_module_24_1.py",
    )
    output = client.execute("cloud-init single --name cc_custom_module_24_1")
    if releases.CURRENT_RELEASE >= releases.PLUCKY:
        assert "The 'get_meta_doc()' function is deprecated" in output
    assert "Hello from module" in output



# This file is part of cloud-init. See LICENSE file for license information.

from unittest import mock

import pytest

from cloudinit.net.ephemeral import EphemeralIPNetwork
from cloudinit.subp import ProcessExecutionError
from cloudinit.url_helper import UrlError
from tests.unittests.helpers import does_not_raise
from tests.unittests.util import MockDistro

M_PATH = "cloudinit.net.ephemeral."


class TestEphemeralIPNetwork:
    @pytest.mark.parametrize(
        "ipv6",
        [
            pytest.param(False, id="no_ipv6"),
            pytest.param(True, id="ipv6"),
        ],
    )
    @pytest.mark.parametrize(
        "ipv4",
        [
            pytest.param(False, id="no_ipv4"),
            pytest.param(True, id="ipv4"),
        ],
    )
    @pytest.mark.parametrize(
        "has_connectivity",
        [
            pytest.param(True, id="has_connectivity"),
            pytest.param(False, id="no_connectivity"),
        ],
    )
    @mock.patch(M_PATH + "contextlib.ExitStack")
    @mock.patch(M_PATH + "EphemeralIPv6Network")
    @mock.patch(M_PATH + "EphemeralDHCPv4")
    def test_stack_order(
        self,
        m_ephemeral_dhcp_v4,
        m_ephemeral_ip_v6_network,
        m_exit_stack,
        has_connectivity,
        ipv4,
        ipv6,
        caplog,
    ):
        interface = object()
        distro = MockDistro()
        with mock.patch(
            M_PATH + "_check_connectivity_to_imds"
        ) as m_check_connectivity_to_imds:
            m_check_connectivity_to_imds.return_value = (
                "http://fake_url" if has_connectivity else None
            )
            with EphemeralIPNetwork(
                distro,
                interface,
                ipv4=ipv4,
                ipv6=ipv6,
                connectivity_urls_data=[{"url": "http://fake_url"}],
            ) as ephemeral_net:
                pass
        # assert that the connectivity check was called if either ipv4 or ipv6
        # is enabled (__enter__ exits early if both are disabled)
        if ipv4 or ipv6:
            m_check_connectivity_to_imds.assert_called_once()
            # check caplog for appropriate messages based on connectivity
            if has_connectivity:
                url = "http://fake_url"
                assert (
                    f"We already have connectivity to IMDS at {url}"
                    ", skipping DHCP." in caplog.text
                )
            else:
                assert (
                    "No connectivity to IMDS, attempting DHCP setup."
                    in caplog.text
                )

        expected_call_args_list = []
        # ipv4 should only be attempted if it is enabled and there is no
        # connectivity before the ephemeral network is brought up
        if ipv4 and not has_connectivity:
            expected_call_args_list.append(
                mock.call(m_ephemeral_dhcp_v4.return_value)
            )
            assert [
                mock.call(
                    distro=distro,
                    iface=interface,
                )
            ] == m_ephemeral_dhcp_v4.call_args_list
        # otherwise, assert that ephemeral_dhcp_v4 was not called
        else:
            assert [] == m_ephemeral_dhcp_v4.call_args_list

        # likewise, ipv6 should only be attempted if it is enabled and there is
        # no connectivity before the ephemeral network is brought up
        if ipv6 and not has_connectivity:
            expected_call_args_list.append(
                mock.call(m_ephemeral_ip_v6_network.return_value)
            )
            assert [
                mock.call(distro, interface)
            ] == m_ephemeral_ip_v6_network.call_args_list
        else:
            assert [] == m_ephemeral_ip_v6_network.call_args_list

        assert (
            expected_call_args_list
            == m_exit_stack.return_value.enter_context.call_args_list
        )
        # if we had to bring up ephemeral ipv6 and we have no ipv4,
        # the state message should reflect that we are using link-local ipv6
        if ipv6 and (not ipv4 and not has_connectivity):
            assert "using link-local ipv6" == ephemeral_net.state_msg

    @pytest.mark.parametrize(
        "m_v4, m_v6, m_context, m_side_effects",
        [
            pytest.param(
                False, True, does_not_raise(), [None, None], id="v6_only"
            ),
            pytest.param(
                True, False, does_not_raise(), [None, None], id="v4_only"
            ),
            pytest.param(
                True,
                True,
                does_not_raise(),
                [ProcessExecutionError, None],
                id="v4_error",
            ),
            pytest.param(
                True,
                True,
                does_not_raise(),
                [None, ProcessExecutionError],
                id="v6_error",
            ),
            pytest.param(
                True,
                True,
                pytest.raises(ProcessExecutionError),
                [
                    ProcessExecutionError,
                    ProcessExecutionError,
                ],
                id="v4_v6_error",
            ),
        ],
    )
    def test_interface_init_failures(
        self, m_v4, m_v6, m_context, m_side_effects, mocker
    ):
        mocker.patch(
            "cloudinit.net.ephemeral.EphemeralDHCPv4"
        ).return_value.__enter__.side_effect = m_side_effects[0]
        mocker.patch(
            "cloudinit.net.ephemeral.EphemeralIPv6Network"
        ).return_value.__enter__.side_effect = m_side_effects[1]
        distro = MockDistro()
        with m_context:
            with EphemeralIPNetwork(distro, "eth0", ipv4=m_v4, ipv6=m_v6):
                pass

    @pytest.mark.parametrize(
        [
            "connectivity_urls_data",
            "has_connectivity",
        ],
        [
            pytest.param(
                [{"url": "http://fake_url"}],
                True,
                id="basic_has_connectivity",
            ),
            pytest.param(
                [{"url": "http://fake_url"}],
                False,
                id="basic_no_connectivity",
            ),
            pytest.param(
                [],
                None,
                id="exits_early_no_urls",
            ),
            pytest.param(
                [{"url": "http://fake_url"}],
                False,
                id="basic_url_error",
            ),
            pytest.param(
                [
                    {"url": "http://fake_url"},
                    {"url": "http://fake_url2", "headers": {"key": "value"}},
                ],
                True,
                id="headers_has_connectivity",
            ),
        ],
    )
    # mock out _do_ipv4 and _do_ipv6
    @mock.patch(
        M_PATH + "EphemeralIPNetwork._perform_ephemeral_network_setup",
        return_value=(True, None),
    )
    @mock.patch(
        M_PATH + "EphemeralIPNetwork._perform_ephemeral_network_setup",
        return_value=(True, None),
    )
    def test_check_connectivity_to_imds(
        self,
        m_do_ipv6,
        m_do_ipv4,
        connectivity_urls_data,
        has_connectivity,
        caplog,
    ):

        def wait_for_url_side_effect(
            urls,
            headers_cb,
            timeout,
            connect_synchronously,
            max_wait,
        ):
            assert urls == [
                url_data["url"] for url_data in connectivity_urls_data
            ]
            for entry in connectivity_urls_data:
                assert headers_cb(entry["url"]) == entry.get("headers")
            if not has_connectivity:
                raise UrlError("fake error")
            return urls[0], b"{}"

        # how wait_for_url is imported in the module:
        # from cloudinit.url_helper import UrlError, wait_for_url

        distro = MockDistro()
        with mock.patch(M_PATH + "wait_for_url") as m_wait_for_url:
            m_wait_for_url.side_effect = wait_for_url_side_effect
            with EphemeralIPNetwork(
                distro,
                "eth0",
                connectivity_urls_data=connectivity_urls_data,
            ):
                pass

        if not connectivity_urls_data:
            assert not m_wait_for_url.called
        elif has_connectivity:
            assert m_wait_for_url.called
        else:
            assert (
                "Failed to reach IMDS without ephemeral network setup: "
                "fake error" in caplog.text
            )
            assert m_wait_for_url.called
        # check caplog for appropriate messages based on connectivity
        if has_connectivity:
            url = [url_data["url"] for url_data in connectivity_urls_data][0]
            assert (
                f"We already have connectivity to IMDS at {url}, "
                "skipping DHCP." in caplog.text
            )
        else:
            assert (
                "No connectivity to IMDS, attempting DHCP setup."
                in caplog.text
            )



import builtins
import glob
import os
import shutil
from pathlib import Path
from unittest import mock

import pytest

from cloudinit import atomic_helper, lifecycle, util
from cloudinit.gpg import GPG
from cloudinit.log import loggers
from tests.hypothesis import HAS_HYPOTHESIS
from tests.unittests.helpers import example_netdev, retarget_many_wrapper


@pytest.fixture
def m_gpg():
    MockGPG = mock.Mock(spec=GPG)
    MockGPG.configure_mock(**{"getkeybyid.return_value": "fakekey"})
    gpg = MockGPG()
    gpg.list_keys = mock.Mock(return_value="<mocked: list_keys>")
    gpg.getkeybyid = mock.Mock(return_value="<mocked: getkeybyid>")

    # to make tests for cc_apt_configure behave, we need the mocked GPG
    # to actually behave like a context manager
    gpg.__enter__ = GPG.__enter__
    gpg.__exit__ = GPG.__exit__
    yield gpg


FS_FUNCS = {
    os.path: [
        ("isfile", 1),
        ("exists", 1),
        ("islink", 1),
        ("isdir", 1),
        ("lexists", 1),
        ("relpath", 1),
    ],
    os: [
        ("listdir", 1),
        ("mkdir", 1),
        ("rmdir", 1),
        ("lstat", 1),
        ("symlink", 2),
        ("stat", 1),
        ("scandir", 1),
    ],
    util: [
        ("write_file", 1),
        ("append_file", 1),
        ("load_binary_file", 1),
        ("load_text_file", 1),
        ("ensure_dir", 1),
        ("chmod", 1),
        ("delete_dir_contents", 1),
        ("del_file", 1),
        ("sym_link", -1),
        ("copy", -1),
    ],
    glob: [
        ("glob", 1),
    ],
    builtins: [
        ("open", 1),
    ],
    atomic_helper: [
        ("write_file", 1),
        ("write_json", 1),
    ],
    shutil: [
        ("rmtree", 1),
    ],
}


@pytest.fixture
def fake_filesystem(mocker, tmpdir):
    """Mocks fs functions to operate under `tmpdir`"""
    # This allows fake_filesystem to be used with production code that
    # creates temporary directories. Functions like TemporaryDirectory()
    # attempt to create a directory under "/tmp" assuming that it already
    # exists, but then it fails because of the retargeting that happens here.
    tmpdir.mkdir("tmp")

    for mod, funcs in FS_FUNCS.items():
        for f, nargs in funcs:
            func = getattr(mod, f)
            trap_func = retarget_many_wrapper(str(tmpdir), nargs, func)
            mocker.patch.object(mod, f, trap_func)
    yield str(tmpdir)


@pytest.fixture(scope="session", autouse=True)
def disable_sysfs_net(tmpdir_factory):
    """Avoid tests which read the underlying host's /syc/class/net."""
    mock_sysfs = f"{tmpdir_factory.mktemp('sysfs')}/"
    with mock.patch(
        "cloudinit.net.get_sys_class_path", return_value=mock_sysfs
    ):
        yield mock_sysfs


@pytest.fixture(scope="class")
def disable_netdev_info(request):
    """Avoid tests which read the underlying host's /syc/class/net."""
    with mock.patch(
        "cloudinit.netinfo.netdev_info", return_value=example_netdev
    ) as mock_netdev:
        yield mock_netdev


@pytest.fixture(autouse=True)
def disable_dns_lookup(request):
    if "allow_dns_lookup" in request.keywords:
        yield
        return

    def side_effect(args, *other_args, **kwargs):
        raise AssertionError("Unexpectedly used util.is_resolvable")

    with mock.patch(
        "cloudinit.util.is_resolvable", side_effect=side_effect, autospec=True
    ):
        yield


@pytest.fixture()
def dhclient_exists():
    with mock.patch(
        "cloudinit.net.dhcp.subp.which",
        return_value="/sbin/dhclient",
        autospec=True,
    ):
        yield


loggers.configure_root_logger()


@pytest.fixture(autouse=True, scope="session")
def disable_root_logger_setup():
    with mock.patch(
        "cloudinit.log.loggers.configure_root_logger", autospec=True
    ):
        yield


@pytest.fixture
def clear_deprecation_log():
    """Clear any deprecation warnings before and after running tests."""
    # Since deprecations are de-duped, the existance (or non-existance) of
    # a deprecation warning in a previous test can cause the next test to
    # fail.
    setattr(lifecycle.deprecate, "log", set())


PYTEST_VERSION_TUPLE = tuple(map(int, pytest.__version__.split(".")))

if PYTEST_VERSION_TUPLE < (3, 9, 0):

    @pytest.fixture
    def tmp_path(tmpdir):
        return Path(tmpdir)


if HAS_HYPOTHESIS:
    from hypothesis import settings  # pylint: disable=import-error

    settings.register_profile("ci", max_examples=1000)
    settings.load_profile(os.getenv("HYPOTHESIS_PROFILE", "default"))



"""test that ds-identify works as expected"""

from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.integration_settings import PLATFORM
from tests.integration_tests.util import (
    verify_clean_boot,
    verify_clean_log,
    wait_for_cloud_init,
)

DATASOURCE_LIST_FILE = "/etc/cloud/cloud.cfg.d/90_dpkg.cfg"
MAP_PLATFORM_TO_DATASOURCE = {
    "lxd_container": "lxd",
    "lxd_vm": "lxd",
    "qemu": "nocloud",
    "ec2": "aws",
    "oci": "oracle",
}


def test_ds_identify(client: IntegrationInstance):
    """Verify that ds-identify works correctly

    Deb packaging often a defines datasource_list with a single datasource,
    which bypasses ds-identify logic. This tests works by removing this file
    and verifying that cloud-init doesn't experience issues.
    """
    assert client.execute(f"rm {DATASOURCE_LIST_FILE}")
    assert client.execute("cloud-init clean --logs")
    client.restart()
    wait_for_cloud_init(client)
    verify_clean_log(client.execute("cat /var/log/cloud-init.log"))
    verify_clean_boot(client)
    assert client.execute("cloud-init status --wait")

    datasource = MAP_PLATFORM_TO_DATASOURCE.get(PLATFORM, PLATFORM)
    cloud_id = client.execute("cloud-id")
    assert cloud_id.ok
    assert datasource == cloud_id.stdout.rstrip()



import logging

from cloudinit import sources

LOG = logging.getLogger(__name__)


class DataSourceNoCacheNetworkOnly(sources.DataSource):
    def _get_data(self):
        LOG.debug("TEST _get_data called")
        return True


datasources = [
    (
        DataSourceNoCacheNetworkOnly,
        (sources.DEP_FILESYSTEM, sources.DEP_NETWORK),
    ),
]


def get_datasource_list(depends):
    return sources.list_from_depends(depends, datasources)



# This file is part of cloud-init. See LICENSE file for license information.

import json
import logging
from typing import Optional

from cloudinit import dmi, sources, url_helper, util

LOG = logging.getLogger(__name__)

METADATA_URLS = ["http://169.254.169.254"]
METADATA_VERSION = 1

CLOUDCIX_DMI_NAME = "CloudCIX"


class DataSourceCloudCIX(sources.DataSource):

    dsname = "CloudCIX"
    # Setup read_url parameters through get_url_params()
    url_retries = 3
    url_timeout_seconds = 5
    url_sec_between_retries = 5

    def __init__(self, sys_cfg, distro, paths):
        super(DataSourceCloudCIX, self).__init__(sys_cfg, distro, paths)
        self._metadata_url = None
        self._net_cfg = None

    def _get_data(self):
        """
        Fetch the user data and the metadata
        """
        try:
            crawled_data = self.crawl_metadata_service()
        except sources.InvalidMetaDataException as error:
            LOG.error(
                "Failed to read data from CloudCIX datasource: %s", error
            )
            return False

        self.metadata = crawled_data["meta-data"]
        self.userdata_raw = util.decode_binary(crawled_data["user-data"])

        return True

    def crawl_metadata_service(self) -> dict:
        md_url = self.determine_md_url()
        if md_url is None:
            raise sources.InvalidMetaDataException(
                "Could not determine metadata URL"
            )

        data = read_metadata(md_url, self.get_url_params())
        return data

    def determine_md_url(self) -> Optional[str]:
        if self._metadata_url:
            return self._metadata_url

        # Try to reach the metadata server
        url_params = self.get_url_params()
        base_url, _ = url_helper.wait_for_url(
            METADATA_URLS,
            max_wait=url_params.max_wait_seconds,
            timeout=url_params.timeout_seconds,
        )
        if not base_url:
            return None

        # Find the highest supported metadata version
        for version in range(METADATA_VERSION, 0, -1):
            url = url_helper.combine_url(
                base_url, "v{0}".format(version), "metadata"
            )
            try:
                response = url_helper.readurl(url, timeout=self.url_timeout)
            except url_helper.UrlError as e:
                LOG.debug("URL %s raised exception %s", url, e)
                continue

            if response.ok():
                self._metadata_url = url_helper.combine_url(
                    base_url, "v{0}".format(version)
                )
                break
            else:
                LOG.debug("No metadata found at URL %s", url)

        return self._metadata_url

    @staticmethod
    def ds_detect():
        return is_platform_viable()

    @property
    def network_config(self):
        if self._net_cfg:
            return self._net_cfg

        if not self.metadata:
            return None
        self._net_cfg = self.metadata["network"]
        return self._net_cfg


def is_platform_viable() -> bool:
    return dmi.read_dmi_data("system-product-name") == CLOUDCIX_DMI_NAME


def read_metadata(base_url: str, url_params):
    """
    Read metadata from metadata server at base_url

    :returns: dictionary of retrieved metadata and user data containing the
              following keys: meta-data, user-data
    :param: base_url: meta data server's base URL
    :param: url_params: dictionary of URL retrieval parameters. Valid keys are
            `retries`, `sec_between` and `timeout`.
    :raises: InvalidMetadataException upon network error connecting to metadata
             URL, error response from meta data server or failure to
             decode/parse metadata and userdata payload.
    """
    md = {}
    leaf_key_format_callback = (
        ("metadata", "meta-data", util.load_json),
        ("userdata", "user-data", util.maybe_b64decode),
    )

    for url_leaf, new_key, format_callback in leaf_key_format_callback:
        try:
            response = url_helper.readurl(
                url=url_helper.combine_url(base_url, url_leaf),
                retries=url_params.num_retries,
                sec_between=url_params.sec_between_retries,
                timeout=url_params.timeout_seconds,
            )
        except url_helper.UrlError as error:
            raise sources.InvalidMetaDataException(
                f"Failed to fetch IMDS {url_leaf}: "
                f"{base_url}/{url_leaf}: {error}"
            )

        if not response.ok():
            raise sources.InvalidMetaDataException(
                f"No valid {url_leaf} found. "
                f"URL {base_url}/{url_leaf} returned code {response.code}"
            )

        try:
            md[new_key] = format_callback(response.contents)
        except json.decoder.JSONDecodeError as exc:
            raise sources.InvalidMetaDataException(
                f"Invalid JSON at {base_url}/{url_leaf}: {exc}"
            ) from exc
    return md


# Used to match classes to dependencies
datasources = [
    (DataSourceCloudCIX, (sources.DEP_FILESYSTEM, sources.DEP_NETWORK)),
]


# Return a list of data sources that match this set of dependencies
def get_datasource_list(depends):
    return sources.list_from_depends(depends, datasources)



# This file is part of cloud-init. See LICENSE file for license information.
"""Datasource for Oracle (OCI/Oracle Cloud Infrastructure)

Notes:
 * This datasource does not support OCI Classic. OCI Classic provides an EC2
   lookalike metadata service.
 * The UUID provided in DMI data is not the same as the meta-data provided
   instance-id, but has an equivalent lifespan.
 * We do need to support upgrade from an instance that cloud-init
   identified as OpenStack.
 * Bare metal instances use iSCSI root, virtual machine instances do not.
 * Both bare metal and virtual machine instances provide a chassis-asset-tag of
   OracleCloud.com.
"""

import base64
import ipaddress
import json
import logging
import time
from typing import Any, Dict, List, NamedTuple, Optional, Tuple

from cloudinit import atomic_helper, dmi, net, sources, util
from cloudinit.distros.networking import NetworkConfig
from cloudinit.net import (
    cmdline,
    ephemeral,
    get_interfaces_by_mac,
    is_netfail_master,
)
from cloudinit.url_helper import wait_for_url

LOG = logging.getLogger(__name__)

BUILTIN_DS_CONFIG = {
    # Don't use IMDS to configure secondary NICs by default
    "configure_secondary_nics": False,
}
CHASSIS_ASSET_TAG = "OracleCloud.com"
IPV4_METADATA_ROOT = "http://169.254.169.254/opc/v{version}/"
IPV6_METADATA_ROOT = "http://[fd00:c1::a9fe:a9fe]/opc/v{version}/"
IPV4_METADATA_PATTERN = IPV4_METADATA_ROOT + "{path}/"
IPV6_METADATA_PATTERN = IPV6_METADATA_ROOT + "{path}/"

V2_HEADERS = {"Authorization": "Bearer Oracle"}

# https://docs.cloud.oracle.com/iaas/Content/Network/Troubleshoot/connectionhang.htm#Overview,
# indicates that an MTU of 9000 is used within OCI
MTU = 9000


class ReadOpcMetadataResponse(NamedTuple):
    version: int
    instance_data: Dict[str, Any]
    vnics_data: Optional[Dict[str, Any]]
    imds_url_used: str


class KlibcOracleNetworkConfigSource(cmdline.KlibcNetworkConfigSource):
    """Override super class to lower the applicability conditions.

    If any `/run/net-*.cfg` files exist, then it is applicable. Even if
    `/run/initramfs/open-iscsi.interface` does not exist.
    """

    def is_applicable(self) -> bool:
        """Override is_applicable"""
        return bool(self._files)


def _ensure_netfailover_safe(network_config: NetworkConfig) -> None:
    """
    Search network config physical interfaces to see if any of them are
    a netfailover master.  If found, we prevent matching by MAC as the other
    failover devices have the same MAC but need to be ignored.

    Note: we rely on cloudinit.net changes which prevent netfailover devices
    from being present in the provided network config.  For more details about
    netfailover devices, refer to cloudinit.net module.

    :param network_config
       A v1 or v2 network config dict with the primary NIC, and possibly
       secondary nic configured.  This dict will be mutated.

    """
    # ignore anything that's not an actual network-config
    if "version" not in network_config:
        return

    if network_config["version"] not in [1, 2]:
        LOG.debug(
            "Ignoring unknown network config version: %s",
            network_config["version"],
        )
        return

    mac_to_name = get_interfaces_by_mac()
    if network_config["version"] == 1:
        for cfg in [c for c in network_config["config"] if "type" in c]:
            if cfg["type"] == "physical":
                if "mac_address" in cfg:
                    mac = cfg["mac_address"]
                    cur_name = mac_to_name.get(mac)
                    if not cur_name:
                        continue
                    elif is_netfail_master(cur_name):
                        del cfg["mac_address"]

    elif network_config["version"] == 2:
        for _, cfg in network_config.get("ethernets", {}).items():
            if "match" in cfg:
                macaddr = cfg.get("match", {}).get("macaddress")
                if macaddr:
                    cur_name = mac_to_name.get(macaddr)
                    if not cur_name:
                        continue
                    elif is_netfail_master(cur_name):
                        del cfg["match"]["macaddress"]
                        del cfg["set-name"]
                        cfg["match"]["name"] = cur_name


class DataSourceOracle(sources.DataSource):

    dsname = "Oracle"
    system_uuid = None
    network_config_sources: Tuple[sources.NetworkConfigSource, ...] = (
        sources.NetworkConfigSource.CMD_LINE,
        sources.NetworkConfigSource.SYSTEM_CFG,
        sources.NetworkConfigSource.DS,
        sources.NetworkConfigSource.INITRAMFS,
    )

    # for init-local stage, we want to bring up an ephemeral network
    perform_dhcp_setup = True

    # Careful...these can be overridden in __init__
    url_max_wait = 30
    url_timeout = 5

    def __init__(self, sys_cfg, *args, **kwargs):
        super(DataSourceOracle, self).__init__(sys_cfg, *args, **kwargs)
        self._vnics_data = None

        self.ds_cfg = util.mergemanydict(
            [
                util.get_cfg_by_path(sys_cfg, ["datasource", self.dsname], {}),
                BUILTIN_DS_CONFIG,
            ]
        )
        self._network_config_source = KlibcOracleNetworkConfigSource()
        self._network_config: dict = {"config": [], "version": 1}

        url_params = self.get_url_params()
        self.url_max_wait = url_params.max_wait_seconds
        self.url_timeout = url_params.timeout_seconds

    def _unpickle(self, ci_pkl_version: int) -> None:
        super()._unpickle(ci_pkl_version)
        if not hasattr(self, "_vnics_data"):
            setattr(self, "_vnics_data", None)
        if not hasattr(self, "_network_config_source"):
            setattr(
                self,
                "_network_config_source",
                KlibcOracleNetworkConfigSource(),
            )
        if not hasattr(self, "_network_config"):
            self._network_config = {"config": [], "version": 1}

    def _has_network_config(self) -> bool:
        return bool(self._network_config.get("config", []))

    @staticmethod
    def ds_detect() -> bool:
        """Check platform environment to report if this datasource may run."""
        return _is_platform_viable()

    def _get_data(self):

        self.system_uuid = _read_system_uuid()

        connectivity_urls_data = (
            {
                "url": IPV4_METADATA_PATTERN.format(
                    version=2, path="instance"
                ),
                "headers": V2_HEADERS,
            },
            {
                "url": IPV4_METADATA_PATTERN.format(
                    version=1, path="instance"
                ),
            },
            {
                "url": IPV6_METADATA_PATTERN.format(
                    version=2, path="instance"
                ),
                "headers": V2_HEADERS,
            },
            {
                "url": IPV6_METADATA_PATTERN.format(
                    version=1, path="instance"
                ),
            },
        )

        if self.perform_dhcp_setup:
            nic_name = net.find_fallback_nic()
            network_context = ephemeral.EphemeralIPNetwork(
                distro=self.distro,
                interface=nic_name,
                ipv6=True,
                ipv4=True,
                connectivity_urls_data=connectivity_urls_data,
            )
        else:
            network_context = util.nullcontext()
        fetch_primary_nic = not self._is_iscsi_root()
        fetch_secondary_nics = self.ds_cfg.get(
            "configure_secondary_nics",
            BUILTIN_DS_CONFIG["configure_secondary_nics"],
        )

        with network_context:
            fetched_metadata = read_opc_metadata(
                fetch_vnics_data=fetch_primary_nic or fetch_secondary_nics,
                max_wait=self.url_max_wait,
                timeout=self.url_timeout,
                metadata_patterns=[
                    IPV6_METADATA_PATTERN,
                    IPV4_METADATA_PATTERN,
                ],
            )

        if not fetched_metadata:
            return False
        # set the metadata root address that worked to allow for detecting
        # whether ipv4 or ipv6 was used for getting metadata
        self.metadata_address = _get_versioned_metadata_base_url(
            url=fetched_metadata.imds_url_used
        )

        data = self._crawled_metadata = fetched_metadata.instance_data
        self._vnics_data = fetched_metadata.vnics_data

        self.metadata = {
            "availability-zone": data["ociAdName"],
            "instance-id": data["id"],
            "launch-index": 0,
            "local-hostname": data["hostname"],
            "name": data["displayName"],
        }

        if "metadata" in data:
            user_data = data["metadata"].get("user_data")
            if user_data:
                self.userdata_raw = base64.b64decode(user_data)
            self.metadata["public_keys"] = data["metadata"].get(
                "ssh_authorized_keys"
            )

        return True

    def check_instance_id(self, sys_cfg) -> bool:
        """quickly check (local only) if self.instance_id is still valid

        On Oracle, the dmi-provided system uuid differs from the instance-id
        but has the same life-span."""
        return sources.instance_id_matches_system_uuid(self.system_uuid)

    def get_public_ssh_keys(self):
        return sources.normalize_pubkey_data(self.metadata.get("public_keys"))

    def _is_iscsi_root(self) -> bool:
        """Return whether we are on a iscsi machine."""
        return self._network_config_source.is_applicable()

    def _get_iscsi_config(self) -> dict:
        return self._network_config_source.render_config()

    @property
    def network_config(self):
        """Network config is read from initramfs provided files

        Priority for primary network_config selection:
        - iscsi
        - imds

        If none is present, then we fall back to fallback configuration.
        """
        if self._has_network_config():
            return self._network_config

        set_primary = False
        # this is v1
        if self._is_iscsi_root():
            self._network_config = self._get_iscsi_config()
        if not self._has_network_config():
            LOG.warning(
                "Could not obtain network configuration from initramfs. "
                "Falling back to IMDS."
            )
            set_primary = True

        set_secondary = self.ds_cfg.get(
            "configure_secondary_nics",
            BUILTIN_DS_CONFIG["configure_secondary_nics"],
        )
        if set_primary or set_secondary:
            try:
                # Mutate self._network_config to include primary and/or
                # secondary VNICs
                self._add_network_config_from_opc_imds(set_primary)
            except Exception:
                util.logexc(
                    LOG,
                    "Failed to parse IMDS network configuration!",
                )

        # we need to verify that the nic selected is not a netfail over
        # device and, if it is a netfail master, then we need to avoid
        # emitting any match by mac
        _ensure_netfailover_safe(self._network_config)

        return self._network_config

    def _add_network_config_from_opc_imds(self, set_primary: bool = False):
        """Generate primary and/or secondary NIC config from IMDS and merge it.

        It will mutate the network config to include the secondary VNICs.

        :param set_primary: If True set primary interface.
        :raises:
        Exceptions are not handled within this function.  Likely
            exceptions are KeyError/IndexError
            (if the IMDS returns valid JSON with unexpected contents).
        """
        if self._vnics_data is None:
            LOG.warning("NIC data is UNSET but should not be")
            return

        if not set_primary and ("nicIndex" in self._vnics_data[0]):
            # TODO: Once configure_secondary_nics defaults to True, lower the
            # level of this log message.  (Currently, if we're running this
            # code at all, someone has explicitly opted-in to secondary
            # VNIC configuration, so we should warn them that it didn't
            # happen.  Once it's default, this would be emitted on every Bare
            # Metal Machine launch, which means INFO or DEBUG would be more
            # appropriate.)
            LOG.warning(
                "VNIC metadata indicates this is a bare metal machine; "
                "skipping secondary VNIC configuration."
            )
            return

        interfaces_by_mac = get_interfaces_by_mac()

        vnics_data = self._vnics_data if set_primary else self._vnics_data[1:]

        # If the metadata address is an IPv6 address

        for index, vnic_dict in enumerate(vnics_data):
            is_primary = set_primary and index == 0
            mac_address = vnic_dict["macAddr"].lower()
            is_ipv6_only = vnic_dict.get(
                "ipv6SubnetCidrBlock", False
            ) and not vnic_dict.get("privateIp", False)
            if mac_address not in interfaces_by_mac:
                LOG.warning(
                    "Interface with MAC %s not found; skipping",
                    mac_address,
                )
                continue
            name = interfaces_by_mac[mac_address]
            if is_ipv6_only:
                network = ipaddress.ip_network(
                    vnic_dict["ipv6Addresses"][0],
                )
            else:
                network = ipaddress.ip_network(vnic_dict["subnetCidrBlock"])

            if self._network_config["version"] == 1:
                if is_primary:
                    if is_ipv6_only:
                        subnets = [{"type": "dhcp6"}]
                    else:
                        subnets = [{"type": "dhcp"}]
                else:
                    subnets = []
                    if vnic_dict.get("privateIp"):
                        subnets.append(
                            {
                                "type": "static",
                                "address": (
                                    f"{vnic_dict['privateIp']}/"
                                    f"{network.prefixlen}"
                                ),
                            }
                        )
                    if vnic_dict.get("ipv6Addresses"):
                        subnets.append(
                            {
                                "type": "static",
                                "address": (
                                    f"{vnic_dict['ipv6Addresses'][0]}/"
                                    f"{network.prefixlen}"
                                ),
                            }
                        )
                interface_config = {
                    "name": name,
                    "type": "physical",
                    "mac_address": mac_address,
                    "mtu": MTU,
                    "subnets": subnets,
                }
                self._network_config["config"].append(interface_config)
            elif self._network_config["version"] == 2:
                # Why does this elif exist???
                # Are there plans to switch to v2?
                interface_config = {
                    "mtu": MTU,
                    "match": {"macaddress": mac_address},
                }
                self._network_config["ethernets"][name] = interface_config

                interface_config["dhcp6"] = is_primary and is_ipv6_only
                interface_config["dhcp4"] = is_primary and not is_ipv6_only
                if not is_primary:
                    interface_config["addresses"] = []
                    if vnic_dict.get("privateIp"):
                        interface_config["addresses"].append(
                            f"{vnic_dict['privateIp']}/{network.prefixlen}"
                        )
                    if vnic_dict.get("ipv6Addresses"):
                        interface_config["addresses"].append(
                            f"{vnic_dict['ipv6Addresses'][0]}/"
                            f"{network.prefixlen}"
                        )
                self._network_config["ethernets"][name] = interface_config


class DataSourceOracleNet(DataSourceOracle):
    perform_dhcp_setup = False


def _is_ipv4_metadata_url(metadata_address: str):
    if not metadata_address:
        return False
    return metadata_address.startswith(IPV4_METADATA_ROOT.split("opc")[0])


def _read_system_uuid() -> Optional[str]:
    sys_uuid = dmi.read_dmi_data("system-uuid")
    return None if sys_uuid is None else sys_uuid.lower()


def _is_platform_viable() -> bool:
    asset_tag = dmi.read_dmi_data("chassis-asset-tag")
    return asset_tag == CHASSIS_ASSET_TAG


def _url_version(url: str) -> int:
    return 2 if "/opc/v2/" in url else 1


def _headers_cb(url: str) -> Optional[Dict[str, str]]:
    return V2_HEADERS if _url_version(url) == 2 else None


def _get_versioned_metadata_base_url(url: str) -> str:
    """
    Remove everything following the version number in the metadata address.
    """
    if not url:
        return url
    if "v2" in url:
        return url.split("v2")[0] + "v2/"
    elif "v1" in url:
        return url.split("v1")[0] + "v1/"
    else:
        raise ValueError("Invalid metadata address: " + url)


def read_opc_metadata(
    *,
    fetch_vnics_data: bool = False,
    max_wait=DataSourceOracle.url_max_wait,
    timeout=DataSourceOracle.url_timeout,
    metadata_patterns: List[str] = [IPV4_METADATA_PATTERN],
) -> Optional[ReadOpcMetadataResponse]:
    """
    Fetch metadata from the /opc/ routes from the IMDS.

    Returns:
        Optional[ReadOpcMetadataResponse]: If fetching metadata fails, None.
            If fetching metadata succeeds, a namedtuple containing:
            - The metadata version as an integer
            - The JSON-decoded value of the instance data from the IMDS
            - The JSON-decoded value of the vnics data from the IMDS if
                `fetch_vnics_data` is True, else None. Alternatively,
                None if fetching metadata failed
            - The url that was used to fetch the metadata.
                This allows for later determining if v1 or v2 endppoint was
                used and whether the IMDS was reached via IPv4 or IPv6.
    """
    urls = [
        metadata_pattern.format(version=version, path="instance")
        for version in [2, 1]
        for metadata_pattern in metadata_patterns
    ]

    LOG.debug("Attempting to fetch IMDS metadata from: %s", urls)
    start_time = time.monotonic()
    url_that_worked, instance_response = wait_for_url(
        urls=urls,
        max_wait=max_wait,
        timeout=timeout,
        headers_cb=_headers_cb,
        sleep_time=0.1,
        connect_synchronously=True,
    )
    if not url_that_worked:
        LOG.warning(
            "Failed to fetch IMDS metadata from any of: %s",
            ", ".join(urls),
        )
        return None
    else:
        LOG.debug(
            "Successfully fetched instance metadata from IMDS at: %s",
            url_that_worked,
        )
    instance_data = json.loads(instance_response.decode("utf-8"))

    # save whichever version we got the instance data from for vnics data later
    metadata_version = _url_version(url_that_worked)

    vnics_data = None
    if fetch_vnics_data:
        # This allows us to go over the max_wait time by the timeout length,
        # but if we were able to retrieve instance metadata, that seems
        # like a worthwhile tradeoff rather than having incomplete metadata.
        vnics_url, vnics_response = wait_for_url(
            urls=[url_that_worked.replace("instance", "vnics")],
            max_wait=max_wait - (time.monotonic() - start_time),
            timeout=timeout,
            headers_cb=_headers_cb,
            sleep_time=0.1,
            connect_synchronously=True,
        )
        if vnics_url:
            vnics_data = json.loads(vnics_response.decode("utf-8"))
            LOG.debug(
                "Successfully fetched vnics metadata from IMDS at: %s",
                vnics_url,
            )
        else:
            LOG.warning("Failed to fetch IMDS network configuration!")
    return ReadOpcMetadataResponse(
        metadata_version,
        instance_data,
        vnics_data,
        url_that_worked,
    )


# Used to match classes to dependencies
datasources = [
    (DataSourceOracle, (sources.DEP_FILESYSTEM,)),
    (
        DataSourceOracleNet,
        (
            sources.DEP_FILESYSTEM,
            sources.DEP_NETWORK,
        ),
    ),
]


# Return a list of data sources that match this set of dependencies
def get_datasource_list(depends):
    return sources.list_from_depends(depends, datasources)


if __name__ == "__main__":

    description = """
        Query Oracle Cloud metadata and emit a JSON object with two keys:
        `read_opc_metadata` and `_is_platform_viable`.  The values of each are
        the return values of the corresponding functions defined in
        DataSourceOracle.py."""

    print(
        atomic_helper.json_dumps(
            {
                "read_opc_metadata": read_opc_metadata(
                    metadata_patterns=(
                        [IPV6_METADATA_PATTERN, IPV4_METADATA_PATTERN]
                    )
                ),
                "_is_platform_viable": _is_platform_viable(),
            }
        )
    )



# This file is part of cloud-init. See LICENSE file for license information.

import os
from collections import namedtuple

import pytest

import cloudinit.settings
from cloudinit.cmd import clean
from cloudinit.distros import Distro
from cloudinit.stages import Init
from cloudinit.util import ensure_dir, sym_link, write_file
from tests.unittests.helpers import mock, wrap_and_call

MyPaths = namedtuple("MyPaths", "cloud_dir")
CleanPaths = namedtuple(
    "CleanPaths",
    ["tmpdir", "cloud_dir", "clean_dir", "log", "output_log"],
)


@pytest.fixture(scope="function")
def clean_paths(tmpdir):
    return CleanPaths(
        tmpdir=tmpdir,
        cloud_dir=tmpdir.join("varlibcloud"),
        clean_dir=tmpdir.join("clean.d"),
        log=tmpdir.join("cloud-init.log"),
        output_log=tmpdir.join("cloud-init-output.log"),
    )


@pytest.fixture(scope="function")
def init_class(clean_paths):
    init = mock.Mock(spec=Init)
    init.paths = MyPaths(cloud_dir=f"{clean_paths.cloud_dir}/")
    init.distro.shutdown_command = Distro.shutdown_command
    init.cfg = {
        "def_log_file": clean_paths.log,
        "output": {"all": f"|tee -a {clean_paths.output_log}"},
    }
    return init


class TestClean:
    def test_remove_artifacts_removes_logs(self, clean_paths, init_class):
        """remove_artifacts removes logs when remove_logs is True."""
        clean_paths.log.write("cloud-init-log")
        clean_paths.output_log.write("cloud-init-output-log")

        assert (
            os.path.exists(clean_paths.cloud_dir) is False
        ), "Unexpected cloud_dir"
        retcode = clean.remove_artifacts(
            init=init_class,
            remove_logs=True,
        )
        assert (
            clean_paths.log.exists() is False
        ), f"Unexpected file {clean_paths.log}"
        assert (
            clean_paths.output_log.exists() is False
        ), f"Unexpected file {clean_paths.output_log}"
        assert 0 == retcode

    def test_remove_net_conf(self, clean_paths, init_class):
        """remove_config removes config files when True."""

        TEST_GEN_NET_CONFIG_FILES = [
            clean_paths.tmpdir.join(conf_file)
            for conf_file in clean.GEN_NET_CONFIG_FILES
        ]
        for conf_path in TEST_GEN_NET_CONFIG_FILES:
            assert conf_path.exists() is False, f"Unexpected {conf_path}"
            ensure_dir(os.path.dirname(conf_path))
            if "*" in conf_path.strpath:
                # Expand glob path to interface-specific eth0 filename
                eth0_path = conf_path.strpath.replace("*", "eth0")
                write_file(
                    eth0_path, f"#generated by cloud-init test {eth0_path}"
                )
            else:
                conf_path.write(f"#generated by cloud-init test {conf_path}")

        with mock.patch(
            "cloudinit.cmd.clean.GEN_NET_CONFIG_FILES",
            [f.strpath for f in TEST_GEN_NET_CONFIG_FILES],
        ):
            retcode = clean.remove_artifacts(
                init_class,
                remove_logs=False,
                remove_config=["network"],
            )

        for conf_path in TEST_GEN_NET_CONFIG_FILES:
            if "*" in conf_path.strpath:
                # Expand glob path to interface-specific eth0 filename
                assert (
                    os.path.exists(conf_path.strpath.replace("*", "eth0"))
                    is False
                ), f"Unexpected file {conf_path.strpath.replace('*', 'eth0')}"
            else:
                assert (
                    conf_path.exists() is False
                ), f"Unexpected file {conf_path}"
        assert 0 == retcode

    def test_remove_ssh_conf(self, clean_paths, init_class):
        """remove_config removes config files when True."""

        TEST_GEN_SSH_CONFIG_FILES = [
            clean_paths.tmpdir.join(conf_file)
            for conf_file in clean.GEN_SSH_CONFIG_FILES
        ]
        for conf_path in TEST_GEN_SSH_CONFIG_FILES:
            assert conf_path.exists() is False, f"Unexpected {conf_path}"
            ensure_dir(os.path.dirname(conf_path))
            conf_path.write(f"#generated by cloud-init\ntouch {conf_path}\n")

        with mock.patch(
            "cloudinit.cmd.clean.GEN_SSH_CONFIG_FILES",
            TEST_GEN_SSH_CONFIG_FILES,
        ):
            retcode = clean.remove_artifacts(
                init_class,
                remove_logs=False,
                remove_config=["ssh_config"],
            )

        for conf_path in TEST_GEN_SSH_CONFIG_FILES:
            assert conf_path.exists() is False, f"Unexpected file {conf_path}"
        assert 0 == retcode

    def test_remove_all_conf(self, clean_paths, init_class):
        """remove_config removes config files when True."""

        TEST_GEN_NET_CONFIG_FILES = [
            clean_paths.tmpdir.join(conf_file)
            for conf_file in clean.GEN_NET_CONFIG_FILES
        ]
        for conf_path in TEST_GEN_NET_CONFIG_FILES:
            assert conf_path.exists() is False, f"Unexpected {conf_path}"
            ensure_dir(os.path.dirname(conf_path))
            conf_path.write(f"#generated by cloud-init\ntouch {conf_path}\n")

        TEST_GEN_SSH_CONFIG_FILES = [
            clean_paths.tmpdir.join(conf_file)
            for conf_file in clean.GEN_SSH_CONFIG_FILES
        ]
        for conf_path in TEST_GEN_SSH_CONFIG_FILES:
            assert conf_path.exists() is False, f"Unexpected {conf_path}"
            ensure_dir(os.path.dirname(conf_path))
            conf_path.write(f"#generated by cloud-init\ntouch {conf_path}\n")

        with mock.patch(
            "cloudinit.cmd.clean.GEN_NET_CONFIG_FILES",
            [f.strpath for f in TEST_GEN_NET_CONFIG_FILES],
        ), mock.patch(
            "cloudinit.cmd.clean.GEN_SSH_CONFIG_FILES",
            TEST_GEN_SSH_CONFIG_FILES,
        ):
            retcode = clean.remove_artifacts(
                init_class,
                remove_logs=False,
                remove_config=["all"],
            )

        for conf_path in TEST_GEN_NET_CONFIG_FILES:
            assert conf_path.exists() is False, f"file {conf_path} exists!"
        for conf_path in TEST_GEN_SSH_CONFIG_FILES:
            assert conf_path.exists() is False, f"file {conf_path} exists!"
        assert 0 == retcode

    def test_keep_net_conf(self, clean_paths, init_class):
        """remove_config removes config files when True."""

        TEST_GEN_NET_CONFIG_FILES = [
            clean_paths.tmpdir.join(conf_file)
            for conf_file in clean.GEN_NET_CONFIG_FILES
        ]
        for conf_path in TEST_GEN_NET_CONFIG_FILES:
            assert conf_path.exists() is False, f"Unexpected {conf_path}"
            ensure_dir(os.path.dirname(conf_path))
            conf_path.write(f"#generated by cloud-init\ntouch {conf_path}\n")

        with mock.patch(
            "cloudinit.cmd.clean.GEN_NET_CONFIG_FILES",
            TEST_GEN_NET_CONFIG_FILES,
        ):
            retcode = clean.remove_artifacts(
                init_class,
                remove_logs=False,
                remove_config=[],
            )

        for conf_path in TEST_GEN_NET_CONFIG_FILES:
            assert conf_path.exists() is True, f"file {conf_path} removed!"
        assert 0 == retcode

    @pytest.mark.allow_all_subp
    def test_remove_artifacts_runparts_clean_d(self, clean_paths, init_class):
        """remove_artifacts performs runparts on CLEAN_RUNPARTS_DIR"""
        ensure_dir(clean_paths.cloud_dir)
        artifact_file = clean_paths.tmpdir.join("didit")
        ensure_dir(clean_paths.clean_dir)
        assert artifact_file.exists() is False, f"Unexpected {artifact_file}"
        clean_script = clean_paths.clean_dir.join("1.sh")
        clean_script.write(f"#!/bin/sh\ntouch {artifact_file}\n")
        clean_script.chmod(mode=0o755)
        with mock.patch.object(
            cloudinit.settings, "CLEAN_RUNPARTS_DIR", clean_paths.clean_dir
        ):
            retcode = clean.remove_artifacts(
                init_class,
                remove_logs=False,
            )
        assert (
            artifact_file.exists() is True
        ), f"Missing expected {artifact_file}"
        assert 0 == retcode

    def test_remove_artifacts_preserves_logs(self, clean_paths, init_class):
        """remove_artifacts leaves logs when remove_logs is False."""
        clean_paths.log.write("cloud-init-log")
        clean_paths.output_log.write("cloud-init-output-log")

        retcode = clean.remove_artifacts(
            init_class,
            remove_logs=False,
        )
        assert 0 == retcode
        assert (
            clean_paths.log.exists() is True
        ), f"Missing expected file {clean_paths.log}"
        assert (
            clean_paths.output_log.exists()
        ), f"Missing expected file {clean_paths.output_log}"

    def test_remove_artifacts_removes_unlinks_symlinks(
        self, clean_paths, init_class
    ):
        """remove_artifacts cleans artifacts dir unlinking any symlinks."""
        dir1 = clean_paths.cloud_dir.join("dir1")
        ensure_dir(dir1)
        symlink = clean_paths.cloud_dir.join("mylink")
        sym_link(dir1.strpath, symlink.strpath)

        with mock.patch.object(
            cloudinit.settings, "CLEAN_RUNPARTS_DIR", clean_paths.clean_dir
        ):
            retcode = clean.remove_artifacts(
                init_class,
                remove_logs=False,
            )
        assert 0 == retcode
        for path in (dir1, symlink):
            assert path.exists() is False, f"Unexpected {path} found"

    def test_remove_artifacts_removes_artifacts_skipping_seed(
        self, clean_paths, init_class
    ):
        """remove_artifacts cleans artifacts dir with exception of seed dir."""
        dirs = [
            clean_paths.cloud_dir,
            clean_paths.cloud_dir.join("seed"),
            clean_paths.cloud_dir.join("dir1"),
            clean_paths.cloud_dir.join("dir2"),
        ]
        for _dir in dirs:
            ensure_dir(_dir)

        with mock.patch.object(
            cloudinit.settings, "CLEAN_RUNPARTS_DIR", clean_paths.clean_dir
        ):
            retcode = clean.remove_artifacts(
                init_class,
                remove_logs=False,
            )
        assert 0 == retcode
        for expected_dir in dirs[:2]:
            assert expected_dir.exists() is True, f"Missing {expected_dir}"
        for deleted_dir in dirs[2:]:
            assert deleted_dir.exists() is False, f"Unexpected {deleted_dir}"

    def test_remove_artifacts_removes_artifacts_removes_seed(
        self, clean_paths, init_class
    ):
        """remove_artifacts removes seed dir when remove_seed is True."""
        dirs = [
            clean_paths.cloud_dir,
            clean_paths.cloud_dir.join("seed"),
            clean_paths.cloud_dir.join("dir1"),
            clean_paths.cloud_dir.join("dir2"),
        ]
        for _dir in dirs:
            ensure_dir(_dir)

        with mock.patch.object(
            cloudinit.settings, "CLEAN_RUNPARTS_DIR", clean_paths.clean_dir
        ):
            retcode = clean.remove_artifacts(
                init_class,
                remove_logs=False,
                remove_seed=True,
            )
        assert 0 == retcode
        assert (
            clean_paths.cloud_dir.exists() is True
        ), f"Missing dir {clean_paths.cloud_dir}"
        for deleted_dir in dirs[1:]:
            assert (
                deleted_dir.exists() is False
            ), f"Unexpected {deleted_dir} dir"

    def test_remove_artifacts_returns_one_on_errors(
        self, clean_paths, init_class, capsys
    ):
        """remove_artifacts returns non-zero on failure and prints an error."""
        ensure_dir(clean_paths.cloud_dir)
        ensure_dir(clean_paths.cloud_dir.join("dir1"))

        with mock.patch(
            "cloudinit.cmd.clean.del_dir", side_effect=OSError("oops")
        ):
            retcode = clean.remove_artifacts(
                init_class,
                remove_logs=False,
            )
        assert 1 == retcode
        _out, err = capsys.readouterr()
        assert (
            f"Error:\nCould not remove {clean_paths.cloud_dir}/dir1: oops\n"
            == err
        )

    def test_handle_clean_args_reboots(self, init_class):
        """handle_clean_args_reboots when reboot arg is provided."""

        called_cmds = []

        def fake_subp(cmd, capture):
            called_cmds.append(cmd)
            return "", ""

        myargs = namedtuple(
            "myargs", "remove_logs remove_seed remove_config reboot machine_id"
        )
        cmdargs = myargs(
            remove_logs=False,
            remove_seed=False,
            remove_config=[],
            reboot=True,
            machine_id=False,
        )
        retcode = wrap_and_call(
            "cloudinit.cmd.clean",
            {
                "subp": {"side_effect": fake_subp},
                "Init": {"return_value": init_class},
            },
            clean.handle_clean_args,
            name="does not matter",
            args=cmdargs,
        )
        assert 0 == retcode
        assert [["shutdown", "-r", "now"]] == called_cmds

    @pytest.mark.parametrize(
        "machine_id,systemd_val",
        (
            pytest.param(True, True, id="machine_id_on_systemd_uninitialized"),
            pytest.param(
                True, False, id="machine_id_non_systemd_removes_file"
            ),
            pytest.param(False, False, id="no_machine_id_param_file_remains"),
        ),
    )
    @mock.patch("cloudinit.cmd.clean.uses_systemd")
    def test_handle_clean_args_removed_machine_id(
        self, uses_systemd, machine_id, systemd_val, clean_paths, init_class
    ):
        """handle_clean_args removes /etc/machine-id when arg is True."""
        uses_systemd.return_value = systemd_val
        myargs = namedtuple(
            "myargs", "remove_logs remove_seed remove_config reboot machine_id"
        )
        cmdargs = myargs(
            remove_logs=False,
            remove_seed=False,
            remove_config=[],
            reboot=False,
            machine_id=machine_id,
        )
        machine_id_path = clean_paths.tmpdir.join("machine-id")
        machine_id_path.write("SOME-AMAZN-MACHINE-ID")
        with mock.patch.object(
            cloudinit.settings, "CLEAN_RUNPARTS_DIR", clean_paths.clean_dir
        ):
            with mock.patch.object(
                cloudinit.cmd.clean, "ETC_MACHINE_ID", machine_id_path.strpath
            ):
                with mock.patch(
                    "cloudinit.cmd.clean.Init", return_value=init_class
                ):
                    assert 0 == clean.handle_clean_args(
                        name="does not matter",
                        args=cmdargs,
                    )
        if systemd_val:
            if machine_id:
                assert "uninitialized\n" == machine_id_path.read()
            else:
                assert "SOME-AMAZN-MACHINE-ID" == machine_id_path.read()
        else:
            assert machine_id_path.exists() is bool(not machine_id)

    def test_status_main(self, clean_paths, init_class):
        """clean.main can be run as a standalone script."""
        clean_paths.log.write("cloud-init-log")
        with pytest.raises(SystemExit) as context_manager:
            wrap_and_call(
                "cloudinit.cmd.clean",
                {
                    "Init": {"return_value": init_class},
                    "sys.argv": {"new": ["clean", "--logs"]},
                },
                clean.main,
            )
        assert 0 == context_manager.value.code
        assert (
            clean_paths.log.exists() is False
        ), f"Unexpected log {clean_paths.log}"



import datetime
import glob
import itertools
import os
import sys

from cloudinit import version
from cloudinit.config.schema import get_schema
from cloudinit.handlers.jinja_template import render_jinja_payload

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath("../../"))
sys.path.insert(0, os.path.abspath("../"))
sys.path.insert(0, os.path.abspath("./"))
sys.path.insert(0, os.path.abspath("."))

# Suppress warnings for docs that aren't used yet
# unused_docs = [
# ]

# General information about the project.
project = "cloud-init"
copyright = f"Canonical Group Ltd, {datetime.date.today().year}"

# -- General configuration ----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = "4.0"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "m2r2",
    "notfound.extension",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.viewcode",
    "sphinxcontrib.datatemplates",
    "sphinxcontrib.mermaid",
    "sphinxcontrib.spelling",
]


# Spelling settings for sphinxcontrib.spelling
# https://docs.ubuntu.com/styleguide/en/
spelling_warning = True

templates_path = ["templates"]
# Uses case-independent spelling matches from doc/rtd/spelling_word_list.txt
spelling_filters = ["spelling.WordListFilter"]
spelling_word_list_filename = "spelling_word_list.txt"

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
version = version.version_string()
release = version

# Set the default Pygments syntax
highlight_language = "yaml"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = False

# Sphinx-copybutton config options: 1) prompt to be stripped from copied code.
# 2) Set to copy all lines (not just prompt lines) to ensure multiline snippets
# can be copied even if they don't contain an EOF line.
copybutton_prompt_text = r"\$ |PS> "
copybutton_prompt_is_regexp = True
copybutton_only_copy_prompt_lines = False

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "furo"
html_theme_options = {
    "light_logo": "logo.png",
    "dark_logo": "logo-dark-mode.png",
    "light_css_variables": {
        "font-stack": "Ubuntu, -apple-system, Segoe UI, Roboto, Oxygen, Cantarell, Fira Sans, Droid Sans, Helvetica Neue, sans-serif",  # noqa: E501
        "font-stack--monospace": "Ubuntu Mono variable, Ubuntu Mono, Consolas, Monaco, Courier, monospace",  # noqa: E501
        "color-foreground-primary": "#111",
        "color-foreground-secondary": "var(--color-foreground-primary)",
        "color-foreground-muted": "#333",
        "color-background-secondary": "#FFF",
        "color-background-hover": "#f2f2f2",
        "color-brand-primary": "#111",
        "color-brand-content": "#06C",
        "color-inline-code-background": "rgba(0,0,0,.03)",
        "color-sidebar-link-text": "#111",
        "color-sidebar-item-background--current": "#ebebeb",
        "color-sidebar-item-background--hover": "#f2f2f2",
        "sidebar-item-line-height": "1.3rem",
        "color-link-underline": "var(--color-background-primary)",
        "color-link-underline--hover": "var(--color-background-primary)",
    },
    "dark_css_variables": {
        "color-foreground-secondary": "var(--color-foreground-primary)",
        "color-foreground-muted": "#CDCDCD",
        "color-background-secondary": "var(--color-background-primary)",
        "color-background-hover": "#666",
        "color-brand-primary": "#fff",
        "color-brand-content": "#06C",
        "color-sidebar-link-text": "#f7f7f7",
        "color-sidebar-item-background--current": "#666",
        "color-sidebar-item-background--hover": "#333",
    },
}

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_static_path = ["static"]
html_css_files = ["css/custom.css", "css/github_issue_links.css"]
html_js_files = [
    "js/github_issue_links.js",
    "js/mermaid_config.js",
    "https://cdn.jsdelivr.net/npm/mermaid@10/dist/mermaid.min.js",
]

html_extra_path = ["googleaf254801a5285c31.html"]

# Make sure the target is unique
autosectionlabel_prefix_document = True
autosectionlabel_maxdepth = 2

# Sphinx-linkcheck config
linkcheck_ignore = [
    r"http://\[fd00:ec2::254.*",
    r"http://instance-data.*",
    r"https://www.scaleway.com/en/developers/api/instance.*",
    r"https://powersj.io.*",
    r"http://169.254.169.254.*",
    r"http://10.10.0.1.*",
]

linkcheck_anchors_ignore_for_url = (
    # Ignore github anchors in rst or md files
    r"https://github.com/.*\.rst",
    r"https://github.com/.*\.md",
    # Ignore github line number anchors in cloud-init and ubuntu-pro-client
    r"https://github.com/canonical/cloud-init.*",
    r"https://github.com/canonical/ubuntu-pro-client.*",
)

new_doc_issue_link = (
    "https://github.com/canonical/cloud-init/issues/new?"
    "labels=documentation%2C+new&projects=&template=documentation.md&"
    "title=%5Bdocs%5D%3A+missing+redirect"
)
docs_url = "https://docs.cloud-init.io"

# Sphinx-copybutton config options:
notfound_body = (
    "<h1>Page not found</h1><p>Sorry we missed you! Our docs have had a"
    " remodel and some deprecated links have changed. Please"
    f" <a href='{new_doc_issue_link}'>file a documentation bug</a> and"
    " we'll fix this redirect.</p><p>"
    f"<a href='{docs_url}'>Back to our homepage hosted at {docs_url}</a></p>"
)

notfound_context = {
    "title": "Page not found",
    "body": notfound_body,
}


def get_types_str(prop_cfg):
    """Return formatted string for all supported config types."""
    types = ""

    # When oneOf present, join each alternative with an '/'
    types += "/".join(
        get_types_str(oneof_cfg) for oneof_cfg in prop_cfg.get("oneOf", [])
    )
    if "items" in prop_cfg:
        types = f"{prop_cfg['type']} of "
        types += get_types_str(prop_cfg["items"])
    elif "enum" in prop_cfg:
        types += f"{'/'.join([f'``{enum}``' for enum in prop_cfg['enum']])}"
    elif "type" in prop_cfg:
        if isinstance(prop_cfg["type"], list):
            types = "/".join(prop_cfg["type"])
        else:
            types = prop_cfg["type"]
    return types


def get_changed_str(prop_name, prop_cfg):
    changed_cfg = {}
    if prop_cfg.get("changed"):
        changed_cfg = prop_cfg
    for oneof_cfg in prop_cfg.get("oneOf", []):
        if oneof_cfg.get("changed"):
            changed_cfg = oneof_cfg
            break
    if changed_cfg:
        with open("templates/property_changed.tmpl", "r") as stream:
            content = "## template: jinja\n" + stream.read()
        return render_jinja_payload(
            content, f"changed_{prop_name}", changed_cfg
        )
    return ""


def get_deprecated_str(prop_name, prop_cfg):
    deprecated_cfg = {}
    if prop_cfg.get("deprecated"):
        deprecated_cfg = prop_cfg
    for oneof_cfg in prop_cfg.get("oneOf", []):
        if oneof_cfg.get("deprecated"):
            deprecated_cfg = oneof_cfg
            break
    if deprecated_cfg:
        with open("templates/property_deprecation.tmpl", "r") as stream:
            content = "## template: jinja\n" + stream.read()
        return render_jinja_payload(
            content, f"deprecation_{prop_name}", deprecated_cfg
        )
    return ""


def render_property_template(prop_name, prop_cfg, prefix=""):
    if prop_cfg.get("description"):
        description = f" {prop_cfg['description']}"
    else:
        description = ""
    description += get_deprecated_str(prop_name, prop_cfg)
    description += get_changed_str(prop_name, prop_cfg)
    jinja_vars = {
        "prefix": prefix,
        "name": prop_name,
        "description": description,
        "types": get_types_str(prop_cfg),
        "prop_cfg": prop_cfg,
    }
    with open("templates/module_property.tmpl", "r") as stream:
        content = "## template: jinja\n" + stream.read()
    return render_jinja_payload(content, f"doc_module_{prop_name}", jinja_vars)


def flatten_schema_refs(src_cfg: dict, defs: dict):
    """Flatten schema: replace $refs in src_cfg with definitions from $defs."""
    if "$ref" in src_cfg:
        reference = src_cfg.pop("$ref").replace("#/$defs/", "")
        # Update the defined references in subschema for doc rendering
        src_cfg.update(defs[reference])
    if "items" in src_cfg:
        if "$ref" in src_cfg["items"]:
            reference = src_cfg["items"].pop("$ref").replace("#/$defs/", "")
            # Update the references in subschema for doc rendering
            src_cfg["items"].update(defs[reference])
        if "oneOf" in src_cfg["items"]:
            for sub_schema in src_cfg["items"]["oneOf"]:
                if "$ref" in sub_schema:
                    reference = sub_schema.pop("$ref").replace("#/$defs/", "")
                    sub_schema.update(defs[reference])
    for sub_schema in itertools.chain(
        src_cfg.get("oneOf", []),
        src_cfg.get("anyOf", []),
        src_cfg.get("allOf", []),
    ):
        if "$ref" in sub_schema:
            reference = sub_schema.pop("$ref").replace("#/$defs/", "")
            sub_schema.update(defs[reference])


def flatten_schema_all_of(src_cfg: dict):
    """Flatten schema: Merge allOf.

    If a schema as allOf, then all of the sub-schemas must hold. Therefore
    it is safe to merge them.
    """
    sub_schemas = src_cfg.pop("allOf", None)
    if not sub_schemas:
        return
    for sub_schema in sub_schemas:
        src_cfg.update(sub_schema)


def render_nested_properties(prop_cfg, defs, prefix):
    prop_str = ""
    prop_types = set(["properties", "patternProperties"])
    flatten_schema_refs(prop_cfg, defs)
    if "items" in prop_cfg:
        prop_str += render_nested_properties(prop_cfg["items"], defs, prefix)
        for alt_schema in prop_cfg["items"].get("oneOf", []):
            if prop_types.intersection(alt_schema):
                prop_str += render_nested_properties(alt_schema, defs, prefix)

    for hidden_key in prop_cfg.get("hidden", []):
        prop_cfg.pop(hidden_key, None)

    # Render visible property types
    for prop_type in prop_types.intersection(prop_cfg):
        for prop_name, nested_cfg in prop_cfg.get(prop_type, {}).items():
            flatten_schema_all_of(nested_cfg)
            flatten_schema_refs(nested_cfg, defs)
            if nested_cfg.get("label"):
                prop_name = nested_cfg.get("label")
            prop_str += render_property_template(prop_name, nested_cfg, prefix)
            prop_str += render_nested_properties(
                nested_cfg, defs, prefix + "  "
            )
    return prop_str


def debug_module_docs(
    module_id: str, mod_docs: dict, debug_file_path: str = None
):
    """Print rendered RST module docs during build.

    The intent is to make rendered RST inconsistencies easier to see when
    modifying jinja template files or JSON schema as white-space and format
    inconsistencies can lead to significant sphinx rendering issues in RTD.

    To trigger this inline print of rendered docs, set the environment
    variable CLOUD_INIT_DEBUG_MODULE_DOC.

    :param module_id: A specific 'cc_*' module name to print rendered RST for,
        or provide 'all' to print out all rendered module docs.
    :param mod_docs: A dict represnting doc metadata for each config module.
        The dict is keyed on config module id (cc_*) and each value is a dict
        with values such as: title, name, examples, schema_doc.
    :param debug_file_path: A specific file to write the rendered RST content.
        When unset,
    """
    from cloudinit.util import load_text_file, load_yaml

    if not module_id:
        return
    if module_id == "all":
        module_ids = mod_docs.keys()
    else:
        module_ids = [module_id]
    rendered_content = ""
    for mod_id in module_ids:
        try:
            data = load_yaml(
                load_text_file(f"../module-docs/{mod_id}/data.yaml")
            )
        except FileNotFoundError:
            continue
        with open("templates/modules.tmpl", "r") as stream:
            tmpl_content = "## template: jinja\n" + stream.read()
            params = {"data": data, "config": {"html_context": mod_docs}}
            rendered_content += render_jinja_payload(
                tmpl_content, "changed_modules_page", params
            )
    if debug_file_path:
        print(f"--- Writing rendered module docs: {debug_file_path} ---")
        with open(debug_file_path, "w") as stream:
            stream.write(rendered_content)
    else:
        print(rendered_content)


def render_module_schemas():
    from cloudinit.importer import import_module

    mod_docs = {}
    schema = get_schema()
    defs = schema.get("$defs", {})

    for mod_path in glob.glob("../../cloudinit/config/cc_*py"):
        mod_name = os.path.basename(mod_path).replace(".py", "")
        mod = import_module(f"cloudinit.config.{mod_name}")
        cc_key = mod.meta["id"]
        mod_docs[cc_key] = {
            "meta": mod.meta,
        }
        if cc_key in defs:
            mod_docs[cc_key]["schema_doc"] = render_nested_properties(
                defs[cc_key], defs, ""
            )
        else:
            mod_docs[cc_key][
                "schema_doc"
            ] = "No schema definitions for this module"
    debug_module_docs(
        os.environ.get("CLOUD_INIT_DEBUG_MODULE_DOC"),
        mod_docs,
        debug_file_path=os.environ.get("CLOUD_INIT_DEBUG_MODULE_DOC_FILE"),
    )
    return mod_docs


html_context = render_module_schemas()
